from matplotlib.backends.backend_pdf import PdfPages
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import json
import numpy as np
import itertools
from argschema import ArgSchemaParser
from nway.schemas import NwayDiagnosticSchema
import os


def cell_experiment_dict(nway_output_path):
    """lookup dict for experiment by cell ID

    Parameters
    ----------
    nway_output_path : str
        path to output json from NwayMatching

    Returns
    -------
    lookup : dict
        dictionary with "cellID": "experimentID"

    """
    with open(nway_output_path, 'r') as f:
        j = json.load(f)
    lookup = {}
    for pw in j['pairwise_results']:
        print(list(pw.keys()))
        for k in ['rejected', 'matches']:
            for pair in pw[k]:
                lookup[pair['fixed']] = pw['fixed_experiment']
                lookup[pair['moving']] = pw['moving_experiment']
        for cellid in pw['unmatched']['fixed']:
            lookup[cellid] = pw['fixed_experiment']
        for cellid in pw['unmatched']['moving']:
            lookup[cellid] = pw['moving_experiment']
    return lookup


def pairwise_transforms(
        nway_output_path, fig=None, subplot_spec=None, fontsize=6):
    """summarize transform parameters and optionally plot

    Parameters
    ----------
    nway_output_path : str
        path to output json from NwayMatching
    fig : :class:`matplotlib.figure.Figure`
        destination figure for plots. No plotting if None
    subplot_spec : :class:`matplotlib.gridspec.SubplotSpec`
        destination SubplotSpec for plots. If None, subplots
        consumer entire figure
    fontsize : int
        fontsize for text in plot

    Returns
    -------
    results : dict
        summarized pairwise transform parameters and experiment ids

    """
    # read the results into lists
    with open(nway_output_path, 'r') as f:
        j = json.load(f)
    pairs = j['pairwise_results']
    props = list(pairs[0]['transform']['properties'].keys())
    results = {}
    for k in props:
        results[k] = [pair['transform']['properties'][k] for pair in pairs]
    results['ids'] = [
            "%d-\n%d" % (
                pair['moving_experiment'],
                pair['fixed_experiment']) for pair in pairs]

    if fig is not None:
        # plot, if a figure is provided
        if subplot_spec is None:
            # these subplots fill entire figure
            outer_plot_grid = gridspec.GridSpec(1, 1, wspace=0.2, hspace=0.2)
            subplot_spec = outer_plot_grid[0]

        inner_plot_grid = gridspec.GridSpecFromSubplotSpec(
                len(props), 1, subplot_spec=subplot_spec,
                wspace=0.1, hspace=0.1)

        x = np.arange(len(results['ids']))
        labels = ['x', 'y']
        for ip, prop in enumerate(props):
            ax = fig.add_subplot(inner_plot_grid[ip])
            ares = np.array(results[prop])
            if len(ares.shape) == 1:
                ares = ares.reshape(-1, 1)
            ares = ares.transpose()
            for iv, v in enumerate(ares):
                ax.plot(x, v, '-o', label=labels[iv])
                ax.set_ylabel(prop, fontsize=fontsize)
            if ares.shape[0] == 2:
                ax.legend(ncol=2, fontsize=fontsize, loc='best', frameon=False)
            ax.set_xticks(x)
            ax.set_xticklabels([])
            if ip == 0:
                ax.set_title(
                        'pairwise transform properties', fontsize=fontsize)
        ax.set_xticklabels(results['ids'], rotation=45, fontsize=fontsize)
        ax.set_xlabel('pairwise experiment IDs', fontsize=fontsize)

    return results


def some_grid(n):
    """specify a roughly square grid with n elements

    Parameters
    ----------
    n : int
        number of elements

    Returns
    -------
    (nrow, ncol) : tuple
        shape for new grid

    """
    nrow = int(np.floor(np.sqrt(n)))
    # NOTE float cast for python 2.7
    ncol = int(np.ceil(float(n) / nrow))
    return (nrow, ncol)


def pairwise_matches(
        nway_output_path, fig=None, subplot_spec=None, fontsize=6):
    """summarize pairwise matches and reject costs

    Parameters
    ----------
    nway_output_path : str
        path to output json from NwayMatching
    fig : :class:`matplotlib.figure.Figure`
        destination figure for plots. No plotting if None
    subplot_spec : :class:`matplotlib.gridspec.SubplotSpec`
        destination SubplotSpec for plots. If None, subplots
        consumer entire figure
    fontsize : int
        fontsize for text in plot

    Returns
    -------
    costs : dict
        costs[pair id] = {
            'matches': list of costs,
            'rejected': list of costs}
    allcosts = dict
        'matches' : concat of all pair matches,
        'rejected' : concate of all pair rejected

    """
    # read the results into lists
    with open(nway_output_path, 'r') as f:
        j = json.load(f)
    pairs = j['pairwise_results']

    costs = {}
    subkeys = ['matches', 'rejected']
    for pair in pairs:
        k = "%d-%d" % (pair['moving_experiment'], pair['fixed_experiment'])
        costs[k] = {}
        for subk in subkeys:
            costs[k][subk] = [m['cost'] for m in pair[subk]]
    allcosts = {}
    for subk in subkeys:
        allcosts[subk] = np.concatenate([c[subk] for c in costs.values()])

    if fig is not None:
        # plot, if a figure is provided
        if subplot_spec is None:
            # these subplots fill entire figure
            outer_plot_grid = gridspec.GridSpec(1, 1, wspace=0.2, hspace=0.2)
            subplot_spec = outer_plot_grid[0]

        inner_plot_grid = gridspec.GridSpecFromSubplotSpec(
                *some_grid(len(costs)), subplot_spec=subplot_spec,
                wspace=0.1, hspace=0.1)

        bins = np.arange(0, 2.1, 0.1)

        for ic, ck in enumerate(costs):
            ax = fig.add_subplot(inner_plot_grid[ic])
            ax.hist(
                    allcosts['matches'],
                    histtype='step',
                    color='k',
                    bins=bins,
                    label='all matches')
            ax.hist(
                    allcosts['rejected'],
                    histtype='step',
                    color='r',
                    bins=bins,
                    label='all rejected')
            ax.hist(
                    costs[ck]['matches'],
                    color='k',
                    alpha=0.5,
                    bins=bins,
                    label='pair matches')
            ax.hist(
                    costs[ck]['rejected'],
                    color='r',
                    alpha=0.5,
                    bins=bins,
                    label='pair rejected')
            ax.legend(loc=2, fontsize=fontsize, title=ck, frameon=False)
            ax.get_legend().get_title().set_fontsize(fontsize)
            ax.set_yscale('log')
            ax.set_xlabel('cost', fontsize=6)
            ax.set_ylabel('match count', fontsize=6)
            rc = inner_plot_grid[ic].get_rows_columns()
            if rc[3] != (rc[0] - 1):
                ax.set_xticks([])
            if rc[5] != 0:
                ax.set_yticks([])

    return costs, allcosts


def nway_matches(
        nway_output_path, fig=None, fontsize=12):
    """summarize pairwise matches and reject costs

    Parameters
    ----------
    nway_output_path : str
        path to output json from NwayMatching
    fig : :class:`matplotlib.figure.Figure`
        destination figure for plots. No plotting if None
    fontsize : int
        fontsize for text in plot

    Returns
    -------
    allnw : dict
        keys are sorted tuples of match sets
        values are :
            n : number of pairwise matches (< max_distance)
            avecost : average pairwise cost

    """

    _, allcosts = pairwise_matches(nway_output_path)

    # read the results into lists
    with open(nway_output_path, 'r') as f:
        j = json.load(f)
    pairs = j['pairwise_results']
    nway = j['nway_matches']

    # make a set of tuples of all pairwise matches and rejected
    # (anything within max_distance)
    # keep the tuples ordered so we can search easily
    allpw = {}
    for pair in pairs:
        for subk in ['matches', 'rejected']:
            pw = {tuple(np.sort([m['moving'], m['fixed']])): m['cost']
                  for m in pair[subk]}
            allpw.update(pw)

    # go through all the nway matches and get the n and mean for each
    # entry
    allnw = {}
    for match in nway:
        k = tuple(match)
        n = 0
        avcost = 0
        for candidate in itertools.combinations(match, 2):
            tsort = tuple(np.sort(candidate))
            if tsort in allpw:
                n += 1
                avcost += allpw[tsort]
        if n != 0:
            allnw[k] = {
                    'n': n,
                    'avecost': avcost / n}

    if fig is not None:
        spec = gridspec.GridSpec(ncols=5, nrows=5, figure=fig)

        ax0 = fig.add_subplot(spec[1:, 0:4])
        ax_histy = fig.add_subplot(spec[1:5, 4:])
        ax_histx = fig.add_subplot(spec[0, 0:4])

        ns = np.array([i['n'] for i in allnw.values()])
        x = np.arange(ns.min(), ns.max() + 1)
        bins = np.arange(0.0, 2.1, 0.1)
        y = []
        cbinned = []
        for ix in x:
            iy = [i['avecost'] for i in allnw.values() if i['n'] == ix]
            y.append(iy)
            h, _ = np.histogram(iy, bins=bins)
            cbinned.append(h)
        cbinned = np.array(cbinned)

        ax0.imshow(
                np.flipud(cbinned),
                extent=[
                    bins.min(),
                    bins.max(),
                    ns.min() - 0.5,
                    ns.max() + 0.5],
                cmap='gray_r',
                aspect='auto')

        xbins = [bins[i:i+2].mean() for i in range(bins.size - 1)]
        ax_histx.bar(
                xbins,
                cbinned.sum(axis=0),
                color='k',
                alpha=0.5,
                width=0.1,
                edgecolor='k')
        ax_histy.barh(
                x,
                cbinned.sum(axis=1),
                color='k',
                alpha=0.5,
                height=1.0,
                edgecolor='k')

        ax0.set_xlabel('cost', fontsize=fontsize)
        ax0.set_ylabel('matches per set', fontsize=fontsize)
        ax_histx.set_xticks([])
        ax_histy.set_yticks([])
        ax_histy.set_xlabel('count', fontsize=fontsize)
        ax_histx.set_ylabel('count', fontsize=fontsize)
        ax_histx.set_title('nway match results', fontsize=fontsize)
        ax_histx.set_xlim(ax0.get_xlim())
        ax_histy.set_ylim(ax0.get_ylim())

    return allnw


def plot_all(nway_output_path, fname=None):
    fs = (12, 8)

    fig1 = plt.figure(clear=True, figsize=fs)
    pairwise_transforms(nway_output_path, fig=fig1)

    fig2 = plt.figure(clear=True, figsize=fs)
    pairwise_matches(nway_output_path, fig=fig2)

    fig3 = plt.figure(clear=True, figsize=fs)
    nway_matches(nway_output_path, fig=fig3)

    if fname is not None:
        p = PdfPages(fname)
        p.savefig(fig1)
        p.savefig(fig2)
        p.savefig(fig3)
        p.close()


class NwayDiagnostics(ArgSchemaParser):
    default_schema = NwayDiagnosticSchema

    def run(self):
        if self.args['use_input_dir']:
            self.args['output_pdf'] = os.path.join(
                    os.path.dirname(self.args['input_json']),
                    os.path.basename(self.args['output_pdf']))

        plot_all(self.args['input_json'], fname=self.args['output_pdf'])


if __name__ == "__main__":  # pragma: no cover
    nd = NwayDiagnostics()
    nd.run()



import nway.image_processing_utils as imu
import pytest
from pathlib import Path
import numpy as np
import PIL.Image
import cv2

TEST_FILE_DIR = Path.resolve(Path(__file__)).parent / "test_files"


@pytest.fixture
def fixed_image_fixture():
    # grab one specific file because failing large translation
    # cases vary image-by-image
    fgen = TEST_FILE_DIR.rglob("avgInt_a1X.png")
    for f in fgen:
        if "ophys_session_775289198" in str(f):
            impath = f
    with PIL.Image.open(impath) as fp:
        image = np.array(fp)
    return image


@pytest.fixture
def flipped_image_fixture(fixed_image_fixture):
    flipped = np.flipud(np.fliplr(fixed_image_fixture))
    return flipped


@pytest.fixture
def moving_image_fixture(fixed_image_fixture, request):
    immoving = cv2.warpPerspective(
            fixed_image_fixture,
            request.param.get("transform"),
            fixed_image_fixture.shape[::-1])
    return immoving, request.param


@pytest.mark.parametrize("preregister", [True, False])
@pytest.mark.parametrize("motion_type, tform_tol, trans_tol", [
    ("MOTION_TRANSLATION", (0.0001, 0.0001), (0.5, 0.0)),
    ("MOTION_EUCLIDEAN", (0.0002, 0.0001), (0.5, 0.0)),
    ("MOTION_AFFINE", (0.0005, 0.005), (0.5, 0.0)),
    ("MOTION_HOMOGRAPHY", (0.005, 0.0001), (0.5, 0.0)),
    ])
@pytest.mark.parametrize(
        "moving_image_fixture",
        [
           # identity
           {"transform": np.array([[1.0, 0.0, 0.0],
                                   [0.0, 1.0, 0.0],
                                   [0.0, 0.0, 1.0]])},
           # x translation
           {"transform": np.array([[1.0, 0.0, 10.0],
                                   [0.0, 1.0, 0.0],
                                   [0.0, 0.0, 1.0]])},
           # x and y translation
           {"transform": np.array([[1.0, 0.0, 10.0],
                                   [0.0, 1.0, -23.0],
                                   [0.0, 0.0, 1.0]])},
           # 5 deg rotation and x and y translation
           {"transform": np.array([[np.cos(0.087), -np.sin(0.087), 10.0],
                                   [np.sin(0.087), np.cos(0.087), -23.0],
                                   [0.0, 0.0, 1.0]])},
           ],
        indirect=["moving_image_fixture"])
def test_register_image_pair(fixed_image_fixture, moving_image_fixture,
                             preregister, motion_type, tform_tol, trans_tol):
    """
    warp an image and see if the register image can recover the transform.
    NOTE: in production, we are using MOTION_EUCLIDEAN as that is the
    expected type of misalignment from the microscope rigs. The underlying
    opencv functions support affine (includes scale) and homography/perspective
    so those options are exposed in case of a new rig and some experimentation.

    tform_tol is (abs, rel) tolerance on 2x2 multiplicative part
    trans_tol is (abs, rel) tolerance of translations
    (abs, rel) are as specified in np.testing.assert_allclose
    As the order of desired transform goes up, it has more trouble
    matching tolerance
    """
    moving_image, param = moving_image_fixture
    expected_transform = param["transform"]
    transform = imu.register_image_pair(
            fixed_image_fixture,
            moving_image,
            500,
            1e-3,
            motion_type,
            5,
            preregister)

    if not np.allclose(expected_transform[0:2, 0:2], np.eye(2)) & \
            (motion_type == 'MOTION_TRANSLATION'):
        # translation can't be expected to solve for rotation
        pass
    else:
        # the multiplicative part of the transform
        np.testing.assert_allclose(transform[0:2, 0:2],
                                   expected_transform[0:2, 0:2],
                                   atol=tform_tol[0], rtol=tform_tol[1])
        # the translation part of the transform
        np.testing.assert_allclose(transform[0:2, 2],
                                   expected_transform[0:2, 2],
                                   atol=trans_tol[0], rtol=trans_tol[1])


@pytest.mark.parametrize("preregister", [True, False])
@pytest.mark.parametrize("motion_type, tform_tol, trans_tol", [
    ("MOTION_EUCLIDEAN", (0.1, 0.1), (0.5, 0.0)),
    ])
@pytest.mark.parametrize(
        "moving_image_fixture, expect_success",
        [
            # x and y translation
            ({"transform": np.array([[1.0, 0.0, 110.0],
                                    [0.0, 1.0, -123.0],
                                    [0.0, 0.0, 1.0]])}, True),
            # small rotation
            ({"transform": np.array([[np.cos(0.025), -np.sin(0.025), 110.0],
                                    [np.sin(0.025), np.cos(0.025), -123.0],
                                    [0.0, 0.0, 1.0]])}, True),
            # the routine can't handle large rotations and translations
            ({"transform": np.array([[np.cos(0.09), -np.sin(0.09), 110.0],
                                    [np.sin(0.09), np.cos(0.09), -123.0],
                                    [0.0, 0.0, 1.0]])}, False),
            ],
        indirect=["moving_image_fixture"])
def test_preregister_for_big_translations(fixed_image_fixture,
                                          moving_image_fixture, preregister,
                                          motion_type, tform_tol,
                                          trans_tol, expect_success):
    """ the preregistration fixes cases that fail for large offsets. These
    are examples of those cases.
    """
    moving_image, param = moving_image_fixture
    expected_transform = param["transform"]
    transform = imu.register_image_pair(
            fixed_image_fixture,
            moving_image,
            500,
            1e-3,
            motion_type,
            5,
            preregister)

    if preregister & expect_success:
        np.testing.assert_allclose(transform[0:2, 0:2],
                                   expected_transform[0:2, 0:2],
                                   atol=tform_tol[0], rtol=tform_tol[1])
        np.testing.assert_allclose(transform[0:2, 2],
                                   expected_transform[0:2, 2],
                                   atol=trans_tol[0], rtol=trans_tol[1])
    else:
        # failure mode is typically a very small translation solution
        assert not np.allclose(transform[0:2, 2],
                               expected_transform[0:2, 2],
                               atol=trans_tol[0], rtol=trans_tol[1])


def test_failure_to_align(fixed_image_fixture, flipped_image_fixture):
    with pytest.raises(
            cv2.error,
            match=(".*Images may be uncorrelated or non-overlapped "
                   "in function 'findTransformECC")):
        imu.register_image_pair(
                fixed_image_fixture,
                flipped_image_fixture,
                50,
                1e-7,
                'MOTION_EUCLIDEAN',
                5,
                True)


@pytest.mark.parametrize("CLAHE_GRID, CLAHE_CLIP",
                         [(8, 2.5), (24, 2.5), (-1, 2.5)])
def test_contrast_adjust(fixed_image_fixture, CLAHE_GRID, CLAHE_CLIP):
    ca_image = imu.contrast_adjust(fixed_image_fixture, CLAHE_GRID, CLAHE_CLIP)
    assert ca_image.shape == fixed_image_fixture.shape
    assert ca_image.dtype == fixed_image_fixture.dtype


@pytest.mark.parametrize("motion_type", [
                         "MOTION_TRANSLATION", "MOTION_EUCLIDEAN",
                         "MOTION_AFFINE", "MOTION_HOMOGRAPHY"])
def test_warp_image(fixed_image_fixture, motion_type):
    tform = np.array([[1.1, 0.1, 12.3],
                      [-0.5, 0.94, -23],
                      [0.0, 0.0, 1.0]])
    warped = imu.warp_image(
            fixed_image_fixture, tform, motion_type, fixed_image_fixture.shape)
    assert warped.shape == fixed_image_fixture.shape
    assert warped.dtype == 'uint8'



import numpy as np
from skimage.transform import AffineTransform
import PIL.Image
import os
import json
from typing import List, Any, Tuple


def create_nice_mask(experiment, output_directory):
    """create a labeled mask from the input json experiment and a translation
    dict to cell IDs, save them to disk, and add their paths to the experiment
    json.

    Parameters
    ----------
    experiment : dict
        output from segmentation
        'ophys_average_intensity_projection_image' : str(path)
        'cell_rois' : dict
        'id' :  str
        'stimulus_name' : str/Null
    output_directory : str
        destination directory for writing masks and translation dicts

    Returns
    -------
    experiment : dict
        same as input with 2 additional key/value pairs
        'nice_mask_path' : str(path)
        'nice_dict_path' : str(path)

    """

    nice_mask, mask_dict = labeled_mask_from_experiment(experiment)

    mask_path = os.path.join(
            output_directory,
            "%d_nice_mask.tif" % experiment['id'])

    nice_mask = [PIL.Image.fromarray(ix) for ix in nice_mask]
    nice_mask[0].save(mask_path, save_all=True, append_images=nice_mask[1:])

    experiment['nice_mask_path'] = mask_path

    dict_path = os.path.join(
            output_directory,
            "%d_nice_dict.json" % experiment['id'])
    full_dict = {
            'experiment': experiment['id'],
            'mask_path': mask_path,
            'mask_dict': mask_dict
            }

    with open(dict_path, 'w') as f:
        json.dump(full_dict, f, indent=2)

    experiment['nice_dict_path'] = dict_path
    # marshmallow 3.0.0rc6 is less forgiving about extra keys around
    # so, pop out the unused extra keys here
    experiment.pop('stimulus_name')

    return experiment


def row_col_from_roi(roi):
    """from an roi dict, get a list of rows and columns
    that are True in the mask_matrix

    Parameters
    ----------
    roi : dict
        x: x (col) offset
        y: y (row) offset
        width: col width
        height: row height
        mask_matrix: nested boolean list h x w

    Returns
    -------
    masked : :class:`numpy.ndarry`
        N x 2 array of [row, col] pairs

    """
    x0 = roi["x"]
    y0 = roi["y"]
    w = roi["width"]
    h = roi["height"]
    coord = np.mgrid[y0:(y0 + h):1, x0:(x0 + w):1]
    mask = np.array(roi['mask_matrix'])
    masked = np.array([coord[i][mask]
                       for i in range(coord.shape[0])]).reshape(2, -1).T
    return masked


def layered_mask_from_rois(rois: List[Any],
                           shape: Tuple[int, int]) -> np.ndarray:
    """makes a (nz, *shape) dimensioned array where each layer ("z")
    has no overlapping ROIs

    Parameters
    ----------
    rois: list
        list of ROIs. The list is the same as "cell_rois" in
        ExperimentSchema
    shape: tuple
        (nrows, ncols) of the destination image

    Returns
    -------
    masks: numpy.ndarray
       dimensions (nz, nrows, ncols)

    """
    def new_layer():
        return np.zeros(shape, dtype=np.uint32)

    masks = [new_layer()]
    for roi in rois:
        tmp = new_layer()
        tmp[roi['y']:(roi['y'] + roi['height']),
            roi['x']:(roi['x'] + roi['width'])] = np.uint32(roi['mask_matrix'])
        tmp *= roi['id']
        bool_tmp = tmp != 0
        placed = False
        for imask in range(len(masks)):
            bool_mask = masks[imask] != 0
            if np.any(bool_tmp & bool_mask):
                # some overlap, move to next layer
                pass
            else:
                # no overlap, add to this layer
                masks[imask] += tmp
                placed = True
                break
        if not placed:
            masks.append(tmp)
    masks = np.array(masks)

    return masks


def labeled_mask_from_experiment(exp):
    """create a labeled mask and to-cell-id dictionary
    from the experiment json

    Parameters
    ----------
    exp : dict
        id : unique experiment id from LIMS
        ophys_average_intensity_projection_image : path to
        intensity image. used to establish lateral size
        of the mask
        cell_rois : list of dict
            x: x (col) offset
            y: y (row) offset
            width: col width
            height: row height
            mask_matrix: nested boolean list h x w
            id : unique cell id, uint32 from LIMS

    Returns
    -------
    relabeled : :class:`numpy.ndarray`
        depth x n x m uint16 mask with new labels
    rdict : dict
        str(label in mask): cell ID

    """
    with PIL.Image.open(
            exp['ophys_average_intensity_projection_image']) as fim:
        im = np.array(fim)
    # layered mask with uint32 ids as intensities
    mask = layered_mask_from_rois(exp['cell_rois'], im.shape)

    # relabel as ordered uint16 and a translating map
    x = np.unique(mask)
    relabeled = np.zeros_like(mask).astype('uint16')
    rdict = {}
    for i, ix in enumerate(x):
        ind = np.nonzero(mask == ix)
        relabeled[ind] = i
        rdict[str(i)] = int(ix)

    # do not list the background
    rdict.pop("0")

    return relabeled, rdict


def calc_first_order_properties(M):
    """human-readable affine properties of a transform matrix

    Parameters
    ----------
    M : :class:`numpy.ndarray`
       3 x 3 augmented affine or projective matrix

    Returns
    -------
    summary : dict
        human-readable dict of transform properties

    """

    summary = {}
    if not np.all(M[2] == np.array([0.0, 0.0, 1.0])):
        summary['warning'] = "affine properties of projective transform"

    tform = AffineTransform(M)
    summary['scale'] = tform.scale
    summary['shear'] = tform.shear
    summary['translation'] = tuple(tform.translation)
    summary['rotation'] = tform.rotation

    return summary


def read_tiff_3d(filename):
    """read in 3D tiff for mask

    Parameters
    ----------
    filename : str
        path to tiff file

    Returns
    -------
    img : :class:`numpy.ndarray`
        depth x n x m image
    """

    with PIL.Image.open(filename) as ims:
        sz = ims.size
        nframes = ims.n_frames
        im = np.zeros((nframes, sz[1], sz[0])).astype('uint16')
        for i, iim in enumerate(PIL.ImageSequence.Iterator(ims)):
            im[i] = np.array(iim)

    return im



import cv2
import numpy as np
import logging

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


def register_image_pair(img_fixed: np.ndarray, img_moving: np.ndarray,
                        maxCount: int, epsilon: float,
                        motion_type: str, gaussFiltSize: int,
                        preregister: bool = True) -> np.ndarray:
    """find the transform that registers two images

    Parameters
    ----------
    img_fixed : numpy.ndarray
        CV_8U (uint8) or CV_32F (float32) fixed image
    img_moving : numpy.ndarray
        CV_8U (uint8) or CV_32F (float32) moving image
    maxCount : int
        passed as maxCount to opencv termination criteria
    epsilon : float
        passed as epsilon to opencv termination criteria
    motion_type : str
        one of the 4 possible motion types for opencv findTransformECC
        see the dictionary cvmotion below for the 4 possible values
    gaussFiltSize : int
        passed to opencv findTransformECC(). An optional value
        indicating size of gaussian blur filter
    preregister: bool
        whether to give a hint to cv2.findTransformECC from cv2.phaseCorrelate

    Returns
    -------
    tform : :class:`numpy.ndarry`
        3 x 3 transformation matrix

    """

    cvmotion = {
            "MOTION_TRANSLATION": cv2.MOTION_TRANSLATION,
            "MOTION_EUCLIDEAN": cv2.MOTION_EUCLIDEAN,
            "MOTION_AFFINE": cv2.MOTION_AFFINE,
            "MOTION_HOMOGRAPHY": cv2.MOTION_HOMOGRAPHY}

    # Define termination criteria
    criteria = (
            cv2.TERM_CRITERIA_EPS | cv2.TERM_CRITERIA_COUNT,
            maxCount,
            epsilon)

    dx = dy = 0.0
    if preregister:
        (dx, dy), _ = cv2.phaseCorrelate(src1=img_fixed.astype('float32'),
                                         src2=img_moving.astype('float32'))

    tform = np.array([[1.0, 0.0, dx],
                      [0.0, 1.0, dy]]).astype('float32')
    if motion_type == 'MOTION_HOMOGRAPHY':
        hrow = np.array([0.0, 0.0, 1.0]).astype('float32')
        tform = np.vstack((tform, hrow))

    try:
        # Run the ECC algorithm. The results are stored in tform
        ccval, tform = cv2.findTransformECC(
                templateImage=img_fixed,
                inputImage=img_moving,
                warpMatrix=tform,
                motionType=cvmotion[motion_type],
                criteria=criteria,
                inputMask=None,
                gaussFiltSize=gaussFiltSize)
    except cv2.error:
        logger.error("failed to align images.")
        raise

    # not all the transforms output 3 x 3
    if tform.shape == (2, 3):
        tform = np.vstack((tform, [0, 0, 1]))

    return tform


def contrast_adjust(image, CLAHE_grid, CLAHE_clip):
    """contrast adjust images

    Parameters
    ----------
    image: numpy.ndaray
        image to pre-process. CV_8UC1 (uint8) or CV_16UC1 (uint16)
    CLAHE_grid : int
        passed as tileGridSize to cv2.createCLAHE. If -1, skipped
    CLAHE_clip : float
        passed as clipLimit to cv2.createCLAHE

    Returns
    -------
    image: numpy.narray
        contrast-adjusted image

    """
    if CLAHE_grid != -1:
        clahe = cv2.createCLAHE(
            clipLimit=CLAHE_clip,
            tileGridSize=(CLAHE_grid, CLAHE_grid))
        image = clahe.apply(src=image)
    return image


def warp_image(image, tform, motion_type, dst_shape):
    """

    Parameters
    ----------
    image: numpy.ndarray
        image to warp. Assumes uint8
    tform: numpy.ndarray
        3x3 transformation matrix
    motion_type: str
        kind of motion. See function `register_image_pair` for possible
        values
    dst_shape: tuple
        desired shape of warped image

    Returns
    -------
    image: numpy.ndarray
        warped image, forced to uint8

    """
    if motion_type == 'MOTION_HOMOGRAPHY':
        warp = cv2.warpPerspective
    else:
        warp = cv2.warpAffine
        if tform.shape[0] == 3:
            tform = tform[0:2]

    image = warp(
            src=image,
            M=tform,
            dsize=dst_shape[::-1],
            flags=cv2.INTER_LINEAR + cv2.WARP_INVERSE_MAP).astype(np.uint8)

    return image



import pytest
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import nway.diagnostics as nwdi
import os
import PyPDF2
import shutil
import json

TEST_FILE_DIR = os.path.join(
        os.path.dirname(os.path.abspath(__file__)),
        'test_files')


@pytest.fixture(scope='module')
def new_output():
    ojpath = os.path.join(
            TEST_FILE_DIR,
            "test0",
            "new_output.json")
    yield ojpath


def test_cell_lookup(new_output):
    cell_lookup = nwdi.cell_experiment_dict(new_output)
    cells = list(cell_lookup.keys())
    with open(new_output, 'r') as f:
        outj = json.load(f)
    for match in outj['nway_matches']:
        for cell_id in match:
            assert cell_id in cells


def test_pairwise_transforms(new_output):
    ptf = nwdi.pairwise_transforms(new_output)
    assert type(ptf) == dict

    fig = plt.figure(1)
    ptf = nwdi.pairwise_transforms(new_output, fig=fig)
    assert type(ptf) == dict

    outer_plot_grid = gridspec.GridSpec(2, 2, wspace=0.2, hspace=0.2)
    subplot_spec = outer_plot_grid[1]
    fig.clf()
    ptf = nwdi.pairwise_transforms(
            new_output, fig=fig, subplot_spec=subplot_spec)
    assert type(ptf) == dict


def test_some_grid():
    for n in range(1, 20):
        (nr, nc) = nwdi.some_grid(n)
        assert (nr * nc) >= n


def test_pairwise_matches(new_output):
    costs, allcosts = nwdi.pairwise_matches(new_output)
    assert type(costs) == dict
    assert type(allcosts) == dict

    fig = plt.figure(1)
    costs, allcosts = nwdi.pairwise_matches(new_output, fig=fig)
    assert type(costs) == dict
    assert type(allcosts) == dict

    outer_plot_grid = gridspec.GridSpec(2, 2, wspace=0.2, hspace=0.2)
    subplot_spec = outer_plot_grid[1]
    fig.clf()
    costs, allcosts = nwdi.pairwise_matches(
            new_output, fig=fig, subplot_spec=subplot_spec)
    assert type(costs) == dict
    assert type(allcosts) == dict


def test_nway_matches(new_output):
    allnw = nwdi.nway_matches(new_output)
    assert type(allnw) == dict

    fig = plt.figure(1)
    allnw = nwdi.nway_matches(new_output, fig=fig)
    assert type(allnw) == dict


def test_plot_all(new_output, tmpdir):
    nwdi.plot_all(new_output)

    output_dir = str(tmpdir.mkdir("diagnostic_test"))
    fname = os.path.join(
            output_dir,
            "output.pdf")

    nwdi.plot_all(new_output, fname)
    assert os.path.isfile(fname)
    PyPDF2.PdfFileReader(fname)


def test_NwayDiagnostics(new_output, tmpdir):
    output_dir = str(tmpdir.mkdir("diagnostic_test"))
    fname = os.path.join(
            output_dir,
            "output2.pdf")

    args = {
            'input_json': new_output,
            'output_pdf': fname,
            'log_level': "DEBUG"
            }

    nd = nwdi.NwayDiagnostics(input_data=args, args=[])
    nd.run()
    assert os.path.isfile(fname)
    PyPDF2.PdfFileReader(fname)

    ninput = os.path.join(
            output_dir,
            "tmpinput.json")

    shutil.copy(new_output, ninput)
    args = {
            'input_json': ninput,
            'output_pdf': "tmp.pdf",
            'use_input_dir': True,
            'log_level': "DEBUG"
            }
    nd = nwdi.NwayDiagnostics(input_data=args, args=[])
    nd.run()

    oname = os.path.join(
            output_dir,
            os.path.basename(args['output_pdf']))
    assert os.path.isfile(oname)
    PyPDF2.PdfFileReader(oname)
